package de.mdelab.sdm.interpreter.code.debug.ui.launcher;

import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.xmi.XMLResource;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PlatformUI;

import de.mdelab.sdm.interpreter.core.AbortInterpretationException;
import de.mdelab.sdm.interpreter.core.debug.SDEclipseInterpreterLauncher;
import de.mdelab.sdm.interpreter.core.debug.SDExecution;
import de.mdelab.sdm.interpreter.core.variables.Variable;

public abstract class SDServerThread<Activity extends EObject, Classifier> extends Thread {
	
	// DB: Issue #1826: Specify schema option for when local ecore metamodel is used
	private static final Map<String, Boolean> defaultSaveOptions = Collections.singletonMap( XMLResource.OPTION_SCHEMA_LOCATION, Boolean.TRUE );
	
	private final Activity activity;
	private final List<Variable<Classifier>> variableBindings;
	private final Integer serverPort;
	private final Integer resourceSetSynchronizerPort;
	private final IProgressMonitor monitor;
	private final PrintStream messageStream;
	private final Map<Resource, URI> instanceModels;
	private final boolean logOutput;
	private final String execTraceUriStr;
	
	private SDEclipseInterpreterLauncher<?, Activity, Classifier> launcher;

	public SDServerThread( 	final Activity p_activity,
							final List<Variable<Classifier>> p_variableBindings,
							final Integer p_serverPort,
							final Integer p_resourceSetSynchronizerPort,
							final IProgressMonitor p_monitor,
							final OutputStream p_messageStream,
							final Map<Resource, URI> p_instanceModels,
							final boolean pb_logOutput,
							final String p_execTraceUriStr ) {
		activity = p_activity;
		variableBindings = p_variableBindings;
		serverPort = p_serverPort;
		resourceSetSynchronizerPort = p_resourceSetSynchronizerPort;
		monitor = p_monitor;
		messageStream = new PrintStream( p_messageStream );
		instanceModels = p_instanceModels;
		logOutput = pb_logOutput;
		execTraceUriStr = p_execTraceUriStr;
		launcher = createLauncher();
	}
	
	protected abstract SDEclipseInterpreterLauncher<?, Activity, Classifier> createLauncher();
	
	@Override
	public void run() {
		try {
			final SDExecution<?, Classifier> execResult = launcher.launch( 	activity, 
																			variableBindings,
																			serverPort,
																			resourceSetSynchronizerPort,
																			logOutput,
																			true, 
																			false, 
																			monitor,
																			new ConsoleLoggerAdapter( messageStream ), 
																			messageStream );
			
			//final Resource instObjRes = ( (EObject) variableBindings.get( 0 ).getValue() ).eResource();
			saveResults( execResult );
		}
		catch ( final AbortInterpretationException p_ex ) {		
		}
		catch( final Throwable p_th ) {
			handleError( p_th );	
		}
	}
	
	private void handleError( final Throwable p_th ) {
		p_th.printStackTrace();
		
		Display.getDefault().asyncExec( new Runnable() {
			
			@Override
			public void run() {
				final Shell shell = getShell();
				
				if ( shell != null ) {
					final String message = p_th.getLocalizedMessage() == null ? "Error when interpreting story diagram. Please check log for more info." : p_th.getLocalizedMessage();
					MessageDialog.openError( shell, "Interpreter Execution Error", message );
				}
			}
		} );
	}
	
	private Shell getShell() {
		final IWorkbenchWindow window = PlatformUI.getWorkbench().getActiveWorkbenchWindow();
		
		if ( window != null ) {
			return window.getShell();
		}
		
		return Display.getDefault().getActiveShell();
	}
	
	private void saveResults( final SDExecution<?, Classifier> p_execRes ) 
	throws IOException {
		final ResourceSet resSet = instanceModels.keySet().iterator().next().getResourceSet();
		
		for ( final Map.Entry<Resource, URI> entry : instanceModels.entrySet() ) {
			final Resource inputRes = entry.getKey();
			final URI outputModelUri = entry.getValue();
			
			if ( outputModelUri == null ) {
				inputRes.save( defaultSaveOptions );
			}
			else {
				final Resource outRes = resSet.createResource( outputModelUri );
				outRes.getContents().addAll( inputRes.getContents() );
				outRes.save( defaultSaveOptions );
			}
		}
	
		if ( !execTraceUriStr.isEmpty() && p_execRes.getExecutionTrace() != null ) {
			final URI execTraceUri = URI.createURI( execTraceUriStr );
			final Resource outRes = resSet.createResource( execTraceUri );
			outRes.getContents().add( p_execRes.getExecutionTrace() );
			outRes.save( null );
		}
	}
	
	private class ConsoleLoggerAdapter extends Logger {
		
		private final PrintStream printStream;

		protected ConsoleLoggerAdapter( final PrintStream p_printStream ) {
			super( "console", null );
			
			printStream = p_printStream;
		}
		
		// FIXME: implement other methods
		
		@Override
	    public void info(String msg) {
			printStream.print( msg );
			printStream.println();
		}
		
		@Override
	    public void severe(String msg) {
			printStream.print( msg );
			printStream.println();
		}
		
		// TODO finish
	}
	
	public boolean isReady() {
		return launcher.isReady();
	}
}
