package de.mdelab.sdm.interpreter.code.debug.ui.launcher.table;

import java.util.Iterator;

import org.eclipse.emf.ecore.provider.EcoreItemProviderAdapterFactory;
import org.eclipse.emf.edit.ui.provider.AdapterFactoryLabelProvider;
import org.eclipse.jface.viewers.ArrayContentProvider;
import org.eclipse.jface.viewers.CellLabelProvider;
import org.eclipse.jface.viewers.EditingSupport;
import org.eclipse.jface.viewers.ILabelProvider;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.TableViewerColumn;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;

/**
 * Table to edit instance models for Eclipse launch configuration
 */
public abstract class TableViewerWrapper<RootObject, RowObject, ColumnType> {
	
	protected TableViewer tableViewer;
	
	protected final ILabelProvider labelProvider;
	
	private final TableViewerComparator<ColumnType> comparator;
	
	private TableData<RootObject, RowObject> tableData = null; 
	
	private Button buttonAdd = null;

	private Button buttonRemove = null;
	
	private final boolean canAdd;

	/**
	 * @param parent
	 */
	public TableViewerWrapper( 	final Composite p_parent,
								final boolean pb_canAdd ) {
		tableData = createTableData();
		labelProvider = new AdapterFactoryLabelProvider( new EcoreItemProviderAdapterFactory() );
		comparator = createTableViewerComparator( labelProvider );
		canAdd = pb_canAdd;
	
		addChildControls( p_parent );
	}
	
	protected abstract TableViewerComparator<ColumnType> createTableViewerComparator( ILabelProvider p_labelProvider );
	
	protected abstract TableData<RootObject, RowObject> createTableData();
	
	public void setEnabled( final boolean pb_enabled ) {
		tableViewer.getTable().setEnabled( pb_enabled );
		
		if ( buttonAdd != null ) {
			buttonAdd.setEnabled( pb_enabled );
		}
		
		if ( buttonRemove != null ) {
			if ( pb_enabled ) {
				updateRemoveButtonEnablement();
			}
			else {
				buttonRemove.setEnabled( pb_enabled );
			}
		}
	}
	
	public TableData<RootObject, RowObject> getTableData() {
		return tableData;
	}
	
	protected boolean buttonAddClicked() {
		return false;
	}

	protected void buttonRemoveClicked() {
		final ISelection selection = tableViewer.getSelection();
		
		if ( selection instanceof IStructuredSelection ) {
			final Iterator<?> iterator = ( (IStructuredSelection) selection ).iterator();
			
			while ( iterator.hasNext() ) {
				@SuppressWarnings("unchecked")
				final RowObject selectedRow = (RowObject) iterator.next();
				getTableData().removeRowObject( selectedRow );
			}
			
			tableViewer.refresh();
		}
	}

	/**
	 * Release resources
	 */
	public void dispose() {
		
		// Tell the label provider to release its resources
		labelProvider.dispose();
	}

	protected Button createButton( 	final Composite p_parent,
									final String p_text,
									final String p_toolTipText ) {
		final Button button = new Button( p_parent, SWT.PUSH );
		button.setFont( p_parent.getFont() );
		button.setText( p_text );
		final GridData gd = new GridData();
		button.setLayoutData( gd );	
	//	setButtonDimensionHint(button);
		button.setLayoutData( new GridData( SWT.RIGHT, SWT.TOP, false, false, 1, 1 ) );
		button.setToolTipText( p_toolTipText );
		
		return button;
	}

	protected void addChildControls( final Composite p_composite ) {
		final GridData gridData = new GridData( GridData.HORIZONTAL_ALIGN_FILL | GridData.FILL_BOTH );
		p_composite.setLayoutData(gridData);
		final int nbCol = canAdd ? 2 : 1;

		GridLayout layout = new GridLayout( nbCol, false );
		layout.marginWidth = 4;
		p_composite.setLayout( layout );

		// Create and setup the TableViewer
		createTableViewer( p_composite );

		if ( canAdd ) {
			final Composite btnCompo = new Composite( p_composite, SWT.NULL );
			final GridLayout btnLayout = new GridLayout( 1, false );
			layout.marginWidth = 4;
			btnCompo.setLayout( btnLayout );
			final GridData btnCompoGridData = new GridData( GridData.HORIZONTAL_ALIGN_FILL | GridData.FILL_BOTH );
			btnCompo.setLayoutData( btnCompoGridData );
			buttonAdd = createButton( btnCompo, "Add...", "Add instance models." );
			buttonAdd.setLayoutData( new GridData( GridData.HORIZONTAL_ALIGN_BEGINNING ) );
			buttonAdd.addSelectionListener( new SelectionListener()	{
				
				@Override
				public void widgetDefaultSelected(SelectionEvent e)	{
					buttonAddClicked();
				}
	
				@Override
				public void widgetSelected(SelectionEvent e) {
					buttonAddClicked();
				}
			} );
			
			buttonRemove = createButton( btnCompo, "Remove", "Remove selected instance models." );
			//buttonRemove.setLayoutData( new GridData( GridData.HORIZONTAL_ALIGN_FILL | GridData.FILL_HORIZONTAL ) );
			buttonRemove.addSelectionListener( new SelectionListener()	{
				
				@Override
				public void widgetDefaultSelected(SelectionEvent e)	{
					buttonRemoveClicked();
				}
	
				@Override
				public void widgetSelected(SelectionEvent e) {
					buttonRemoveClicked();
				}
			} );
			
			tableViewer.addSelectionChangedListener( new ISelectionChangedListener() {
				
				@Override
				public void selectionChanged( final SelectionChangedEvent p_event) {
					final ISelection selection = p_event.getSelection();
					
					updateRemoveButtonEnablement( selection );
				}
			} );
		}
	}
	
	protected void updateRemoveButtonEnablement() {
		updateRemoveButtonEnablement( tableViewer.getSelection() );
	}
	
	private void updateRemoveButtonEnablement( final ISelection p_selection ) {
		if ( buttonRemove != null ) {
			buttonRemove.setEnabled( !p_selection.isEmpty() );
		}
	}

	public void addViewerListener( final ITableViewerListener<RowObject> p_listener ) {
		tableData.addTableViewerListener( p_listener );
	}
	
	public boolean removeViewerListener( final ITableViewerListener<RowObject> p_listener ) {
		return tableData.removeTableViewerListener( p_listener ) ;
	}
	
	protected abstract void createColumns();
	
	protected TableViewerColumn createViewerColumn( final int pi_width,
													final String p_colHeader,
													final CellLabelProvider p_labelProvider,
													final EditingSupport p_editingSupport,
													final ColumnType p_colType ) {
		final TableViewerColumn columnViewer = new TableViewerColumn( tableViewer, SWT.NONE );
		columnViewer.setLabelProvider( p_labelProvider );
		columnViewer.setEditingSupport( p_editingSupport );

		final TableColumn column = columnViewer.getColumn();
		column.setWidth( pi_width );
		column.setText( p_colHeader );
		column.setResizable( true );
	    column.setMoveable( true );
	    column.addSelectionListener( createSelectionAdapter( column, p_colType ) );
	    
		return columnViewer;
	}
	
	private SelectionAdapter createSelectionAdapter(	final TableColumn p_column,
		      											final ColumnType p_colType ) {
		final SelectionAdapter selectionAdapter = new SelectionAdapter() {
			@Override
		    public void widgetSelected(SelectionEvent e) {
				comparator.setColumn( p_colType );
		        final int dir = comparator.getDirection();
		        tableViewer.getTable().setSortDirection( dir );
		        tableViewer.getTable().setSortColumn( p_column );
		        tableViewer.refresh();
		      }
		};
		
		return selectionAdapter;
	}
 
	private void createTableViewer( final Composite p_control ) {
		final int style = SWT.MULTI | SWT.BORDER | SWT.H_SCROLL | SWT.V_SCROLL | SWT.FULL_SELECTION | SWT.HIDE_SELECTION;
		tableViewer = new TableViewer( p_control, style );
		createColumns();
		final Table table = tableViewer.getTable();
		table.setLinesVisible( true );
		table.setHeaderVisible( true );
		table.setLayoutData( new GridData( GridData.HORIZONTAL_ALIGN_FILL | GridData.FILL_BOTH ) );
		tableViewer.setContentProvider( new ArrayContentProvider() );
		tableViewer.setComparator( comparator );
	}
}